/* Copyright (c) 2002,2003 Sam Trenholme
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

/* This code is based on the aeshash.pdf file, written by Bram Cohen and
   Ben Laurie.  The code currently uses the 128-bit key and block size of
   MaraDNS's hasher, making a 128-bit hash */

/* Note that you need to change the cipher we call to change these   
   constants */
#define HASH_BITS 128
#define HASH_BYTES (HASH_BITS / 8)

/* Buffered read */
#define BUF_MAX 4096

#include "rng/rng-api-fst.h"
#include <unistd.h>
#include <fcntl.h>

/* Routine that runs the compression funciton of our hash.
   Input: Plaintext we wish to compress (which is HASH_BYTES long),
          current state of the hash
   Output: The current state of the hash is modified
           1 on success, -1 on fail 
*/

int hash_compress(char *input, char *state) {
    BYTE r_inBlock[HASH_BYTES + 1], r_outBlock[HASH_BYTES + 1],
         r_binKey[HASH_BYTES + 1];
    keyInstance r_keyInst;
    cipherInstance r_cipherInst;
    int counter;

#ifdef DEBUG
    for(counter = 0; counter < HASH_BYTES; counter++) {
        if(input[counter] >= 32 && input[counter] < 128) 
            printf("%c",input[counter] & 0xff); 
        else
	    printf("¿");
        }
    printf("\n");
#endif /* DEBUG */

    /* Copy over the input of the hash to the key for the cipher */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        r_binKey[counter] = input[counter];
	}

    /* Copy over the state and make it the "plaintext" of the cipher */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        r_inBlock[counter] = state[counter];
	}

    /* Prepare the encryption */
    if(makeKey(&r_keyInst, DIR_ENCRYPT, HASH_BITS, r_binKey) != 1) {
        return -1;
	}

    if(cipherInit(&r_cipherInst, MODE_ECB, NULL) != 1) {
        return -1;
	}

    /* Perform the encryption */
    if(blockEncrypt(&r_cipherInst, &r_keyInst, r_inBlock, HASH_BITS,
                    r_outBlock) != HASH_BITS) {
        return -1;
	}

    /* XOR the ciphertext with the current hash state */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        r_outBlock[counter] ^= state[counter];
	}

    /* Make the modified ciphertext the new state */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        state[counter] = r_outBlock[counter];
	}

    return 1;
    } 

/* This function will hash a convential C NULL-terminated string.
   The digest is stored (in hex) as a string which this function 
   allocates */

char *hex_mhash(char *text) {
    char *state, *input, *output;
    int counter, len;
    int readed = 0;

    if((state = (char *)malloc(HASH_BYTES + 2)) == NULL) {
    	return 0;
	}
    if((input = (char *)malloc(HASH_BYTES + 2)) == NULL) {
    	return 0;
	}
    if((output = (char *)malloc(HASH_BYTES * 2 + 2)) == NULL) {
        free(state);
	return 0;
	}

    /* Initialize the state */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        state[counter] = 0xff;
	}
   
    for(counter = 0;;counter +=16) {
        for(readed = 0; readed < 16; readed++) {
	    input[readed] = *(text + counter + readed);
	    if(input[readed] == '\0')
	        break;
		}
        if(input[readed] == '\0' && readed < 16)
	    break;
        hash_compress(input,state);
	}

    len = counter + readed;

    if(readed >= HASH_BYTES / 2) {
        while(readed < HASH_BYTES) {
	    input[readed] = 0;
	    readed++;
	    }
        hash_compress(input,state);
	readed = 0;
	}
    
    while(readed < HASH_BYTES - 4) {
        input[readed] = 0;
	readed++;
	}

    /* We count bits, not bytes (as specified in the aeshash pdf document),
       so multiply len times eight */
    len *= 8;
    
    /* Pad the length to the hash input */
    for(counter = 3; counter >= 0; counter--) {
        input[readed] = (len >> (counter * 8)) & 0xff;
	readed++;
	}
    /* Pad that block */
    hash_compress(input,state);
    /* One last has_compress for security reasons */
    hash_compress(state,state);

    /* Convert the output into an ASCII hex string */
    for(counter = 0; counter < HASH_BYTES; counter++) {
        int nibble1, nibble2;
	nibble1 = state[counter] & 0xf0;
	nibble1 >>= 4;
	nibble2 = state[counter] & 0x0f;

        if(nibble1 <= 9) {
	    *(output + counter * 2) = '0' + nibble1;
	    }
        else {
	    *(output + counter * 2) = ('a' - 10) + nibble1;
	    }
        if(nibble2 <= 9) {
	    *(output + 1 + counter * 2) = '0' + nibble2;
	    }
        else {
	    *(output + 1 + counter * 2) = ('a' - 10) + nibble2;
	    }
        }

    /* Free memory and return value */
    free(input);
    free(state);
    return output;
    }

/* The main routine.  This reads a file specified on the command line,
   then makes a hash out of that file. */

main(int argc, char **argv) {

    /* Check the command line argument */
    if(argc != 2) {
        if(argc >= 1) {
            printf("Usage: %s {hash input}\n",argv[0]);
	    exit(1);
	    }
        else {
	    printf("Usage: <this program> {hash input}\n");
	    exit(2);
	    }
        }

    printf("%s %s\n",hex_mhash(argv[1]),argv[1]);

    }
