/* $OpenBSD: blowfish.c,v 1.8 1998/03/04 00:34:17 deraadt Exp $ */
/*
 * Blowfish block cipher for OpenBSD
 * Copyright 1997 Niels Provos <provos@physnet.uni-hamburg.de>
 * All rights reserved.
 *
 * Implementation advice by David Mazieres <dm@lcs.mit.edu>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This code is derived from section 14.3 and the given source
 * in section V of Applied Cryptography, second edition.
 * Blowfish is an unpatented fast block cipher designed by
 * Bruce Schneier.
 */

/* NOTICE NOTICE NOTICE
 * This is *not* the Blowfish cipher designed by Bruce Schneier
 * This is, instead, a 32-bit varient of it which Bruce Schneier
 * points out is "not suggested for actual implementation."
 *
 * In other words, should this cipher prove insecure, it is my
 * responsibility, since I went directly against Bruce's wishes in
 * using this varient in a real-world cryptosystem.
 * 
 * See this reference:
 * B. Schneier, Description of a New Variable-Length Key, 64-Bit 
 * Block Cipher (Blowfish), Fast Software Encryption, Cambridge 
 * Security Workshop Proceedings (December 1993), Springer-Verlag,  
 * 1994, pp. 191-204. 
 *
 * On the web (look up at http://www.counterpane.com/biblio/)
 */


#ifdef TESTING
#include <stdio.h>		/* used for debugging */
#include <string.h>
#endif

#include <sys/types.h>
#include "blf.h"

#undef inline
#ifdef __GNUC__
#define inline __inline
#else				/* !__GNUC__ */
#define inline
#endif				/* !__GNUC__ */

/* Function for Feistel Networks */

#define F(bc, x) ((((bc)->S[0][((x) & 0xF000) >> 12] \
		    + (bc)->S[1][((x) &0x0F00 ) >> 8]) \
		   ^ (bc)->S[2][((x) & 0x00F0) >> 4]) \
		  + (bc)->S[3][(x) & 0x000F])

#define BLFRND(bc,i,j,n) (i ^= F(bc,j) ^ (bc)->P[n])

void
Blowfish_encipher(c, xl, xr)
	blf_ctx *c;
	u_int16_t *xl;
	u_int16_t *xr;
{
	u_int16_t Xl;
	u_int16_t Xr;

	Xl = *xl;
	Xr = *xr;

	Xl ^= c->P[0];
	BLFRND(c, Xr, Xl, 1); BLFRND(c, Xl, Xr, 2);
	BLFRND(c, Xr, Xl, 3); BLFRND(c, Xl, Xr, 4);
	BLFRND(c, Xr, Xl, 5); BLFRND(c, Xl, Xr, 6);
	BLFRND(c, Xr, Xl, 7); BLFRND(c, Xl, Xr, 8);
	BLFRND(c, Xr, Xl, 9); BLFRND(c, Xl, Xr, 10);
	BLFRND(c, Xr, Xl, 11); BLFRND(c, Xl, Xr, 12);
	BLFRND(c, Xr, Xl, 13); BLFRND(c, Xl, Xr, 14);
	BLFRND(c, Xr, Xl, 15); BLFRND(c, Xl, Xr, 16);

	*xl = Xr ^ c->P[17];
	*xr = Xl;
}

void
Blowfish_decipher(c, xl, xr)
	blf_ctx *c;
	u_int16_t *xl;
	u_int16_t *xr;
{
	u_int16_t Xl;
	u_int16_t Xr;

	Xl = *xl;
	Xr = *xr;

	Xl ^= c->P[17];
	BLFRND(c, Xr, Xl, 16); BLFRND(c, Xl, Xr, 15);
	BLFRND(c, Xr, Xl, 14); BLFRND(c, Xl, Xr, 13);
	BLFRND(c, Xr, Xl, 12); BLFRND(c, Xl, Xr, 11);
	BLFRND(c, Xr, Xl, 10); BLFRND(c, Xl, Xr, 9);
	BLFRND(c, Xr, Xl, 8); BLFRND(c, Xl, Xr, 7);
	BLFRND(c, Xr, Xl, 6); BLFRND(c, Xl, Xr, 5);
	BLFRND(c, Xr, Xl, 4); BLFRND(c, Xl, Xr, 3);
	BLFRND(c, Xr, Xl, 2); BLFRND(c, Xl, Xr, 1);

	*xl = Xr ^ c->P[0];
	*xr = Xl;
}

void
Blowfish_initstate(c)
	blf_ctx *c;
{

/* P-box and S-box tables initialized with digits of Pi */

	const blf_ctx initstate =

	{ {
		{
			0xd131, 0x98df, 0x2ffd, 0xd01a,
			0xb8e1, 0x6a26, 0xba7c, 0xf12c,
			0x24a1, 0xb391, 0x0801, 0x858e,
			0x6369, 0x7157, 0xa458, 0xf493,
		},
		{
			0x4b7a, 0xb5b3, 0xdb75, 0xc419,
			0xad6e, 0x49a7, 0x9cee, 0x8fed,
			0xecaa, 0x699a, 0x5664, 0xc2b1,
			0x1936, 0x7509, 0xa059, 0xe418,
		},
		{
			0xe93d, 0x9481, 0xf64c, 0x9469,
			0x4115, 0x7602, 0xbcf4, 0xd4a2,
			0xd408, 0x3320, 0x43b7, 0x5000,
			0x1e39, 0x9724, 0x1421, 0xbf8b,
		},
		{
			0x3a39, 0xd3fa, 0xabc2, 0x5ac5,
			0x5cb0, 0x4fa3, 0xd382, 0x99bc,
			0xd511, 0xbf0f, 0xd62d, 0xc700,
			0xb78c, 0x21a1, 0xb26e, 0x6a36,
		}
	},
	{
		0x243f, 0x85a3, 0x1319, 0x0370,
		0xa409, 0x299f, 0x082e, 0xec4e,
		0x4528, 0x38d0, 0xbe54, 0x34e9,
		0xc0ac, 0xc97c, 0x3f84, 0xb547,
		0x9216, 0x8979
	} };

	*c = initstate;

}

#ifdef __STDC__
u_int16_t
Blowfish_stream2word(const u_int8_t *data, u_int16_t databytes, u_int16_t *current)
#else
u_int16_t
Blowfish_stream2word(data, databytes, current)
	const u_int8_t *data;
	u_int16_t databytes;
	u_int16_t *current;
#endif
{
	u_int8_t i;
	u_int16_t j;
	u_int16_t temp;

	temp = 0x0000;
	j = *current;

	for (i = 0; i < 2; i++, j++) {
		if (j >= databytes)
			j = 0;
		temp = (temp << 8) | data[j];
	}

	*current = j;
	return temp;
}

#if __STDC__
void
Blowfish_expand0state(blf_ctx *c, const u_int8_t *key, u_int16_t keybytes)
#else
void
Blowfish_expand0state(c, key, keybytes)
	blf_ctx *c;
	const u_int8_t *key;
	u_int16_t keybytes;
#endif
{
	u_int16_t i;
	u_int16_t j;
	u_int16_t k;
	u_int16_t temp;
	u_int16_t datal;
	u_int16_t datar;

	j = 0;
	for (i = 0; i < BLF_N + 2; i++) {
		/* Extract 4 int8 to 1 int32 from keystream */
		temp = Blowfish_stream2word(key, keybytes, &j);
		c->P[i] = c->P[i] ^ temp;
	}

	j = 0;
	datal = 0x0000;
	datar = 0x0000;
	for (i = 0; i < BLF_N + 2; i += 2) {
		Blowfish_encipher(c, &datal, &datar);

		c->P[i] = datal;
		c->P[i + 1] = datar;
	}

	for (i = 0; i < 4; i++) {
		for (k = 0; k < 16; k += 2) {
			Blowfish_encipher(c, &datal, &datar);

			c->S[i][k] = datal;
			c->S[i][k + 1] = datar;
		}
	}
}

#if __STDC__
void
blf_key(blf_ctx *c, const u_int8_t *k, u_int16_t len)
#else
void
blf_key(c, k, len)
	blf_ctx *c;
	const u_int8_t *k;
	u_int16_t len;
#endif
{
	/* Initalize S-boxes and subkeys with Pi */
	Blowfish_initstate(c);

	/* Transform S-boxes and subkeys with key */
	Blowfish_expand0state(c, k, len);
}

#if __STDC__
void
blf_enc(blf_ctx *c, u_int16_t *data, u_int16_t blocks)
#else
void
blf_enc(c, data, blocks)
	blf_ctx *c;
	u_int16_t *data;
	u_int16_t blocks;
#endif
{
	u_int16_t *d;
	u_int16_t i;

	d = data;
	for (i = 0; i < blocks; i++) {
		Blowfish_encipher(c, d, d + 1);
		d += 2;
	}
}

#if __STDC__
void
blf_dec(blf_ctx *c, u_int16_t *data, u_int16_t blocks)
#else
void
blf_dec(c, data, blocks)
	blf_ctx *c;
	u_int16_t *data;
	u_int16_t blocks;
#endif
{
	u_int16_t *d;
	u_int16_t i;

	d = data;
	for (i = 0; i < blocks; i++) {
		Blowfish_decipher(c, d, d + 1);
		d += 2;
	}
}
#ifdef TESTING
void
report(u_int16_t data[], u_int16_t len)
{
	u_int16_t i;
	for (i = 0; i < len; i += 2)
		printf("Block %0hd: %04lx %04lx.\n",
		    i / 2, data[i], data[i + 1]);
}
void
main(void)
{

	blf_ctx c;
	char    key[] = "";
	char    key2[] = "abcdefghijklmnopqrstuvwxyz";

	u_int16_t data[10];
	u_int16_t data2[] =
	{0x424c, 0x4649};

	u_int16_t i;

	/* First test */
	for (i = 0; i < 10; i++)
		data[i] = i;

	blf_key(&c, (u_int8_t *) key, 5);
	blf_enc(&c, data, 5);
	blf_dec(&c, data, 1);
	blf_dec(&c, data + 2, 4);
	printf("Should read as 0 - 9.\n");
	report(data, 10);

	/* Second test */
	blf_key(&c, (u_int8_t *) key2, strlen(key2));
	blf_enc(&c, data2, 1);
	printf("\nShould read as: 0x424c 0x4649.\n");
	report(data2, 2);
	blf_dec(&c, data2, 1);
	report(data2, 2);
}
#endif
